#include "../../include/ctx.h"
#include "../../include/pool.h"
#include "../../include/util.h"
#include "../../include/error.h"
#include "../../include/thpool.h"

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <libgen.h>
#include <string.h>

lm_pool_t *lm_ctx_pool_add(lm_ctx_t *ctx, char *name, char *url) {
  if(NULL == name){
    lm_error_set(LM_ERR_ArgNULL);
    return NULL;
  }

  lm_pool_t *pool = lm_pool_new(name, url);
  if (NULL == pool)
    return NULL;

  pdebug(__func__, "adding new pool: %s (%s://%s:%d%s)", pool->name, pool->url.protocol, pool->url.host, pool->url.port, pool->url.path);

  if(NULL != ctx->data){
    char pools_dir[strlen(ctx->data)+strlen(pool->name)+10];
    join(pools_dir, ctx->data, "pools");

    if(!mkdir_ifnot(pools_dir)){
      lm_error_set(LM_ERR_CtxDataFailMkdir);
      lm_pool_free(pool);
      return NULL;
    }

    bzero(pools_dir, sizeof(pools_dir));
    join_multiple(pools_dir,  ctx->data, "pools", pool->name);

    if(!lm_pool_path_set_dir(pool, pools_dir)){
      // error is set by set_dir
      lm_pool_free(pool);
      return NULL;
    }
  }

  if (NULL == ctx->pools) {
    ctx->pools = pool;
    return pool;
  }

  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (NULL == cur->next) {
      cur->next = pool;
      return pool;
    }
    cur = cur->next;
  }

  // if we somehow get here, something very wrong has happend
  assert(true);

  lm_pool_free(pool);
  return NULL;
}

void lm_ctx_pool_clear(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools, *prev = NULL;
  while (NULL != cur) {
    prev = cur;
    cur  = cur->next;
    lm_pool_free(prev);
  }

  ctx->pools = NULL;
}

void lm_ctx_pool_test(lm_ctx_t *ctx, lm_ctx_pool_callback_t callback, void *data) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    lm_pool_test(cur);

    if (!cur->available)
      pdebug(__func__, "(%s) not avaliable: %s", cur->name, lm_strerror());

    if(NULL != callback)
      if(!callback(ctx, cur, cur->available, data))
        break;

    cur = cur->next;
  }
}

lm_pool_t *lm_ctx_pool_find(lm_ctx_t *ctx, char *name) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (eq(cur->name, name))
      return cur;
    cur = cur->next;
  }
  return NULL;
}

lm_pool_t *lm_ctx_pool_by_url(lm_ctx_t *ctx, char *host, char *path) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (eq(cur->url.host, host) && eq(cur->url.path, path))
      return cur;
    cur = cur->next;
  }
  return NULL;
}

void lm_ctx_pool_get_info(lm_ctx_t *ctx, bool allow_update, bool force_update, lm_ctx_pool_callback_t callback, void *data){
  lm_pool_t *cur = ctx->pools;
  bool status = false;

  while(NULL != cur){
    if(lm_pool_path_is_empty(cur)){
      pdebug(__func__, "(%s) failed to load info, pool paths are empty", cur->name);
      goto next;
    }

    if(force_update)
      goto update;

    if(!lm_pool_info_load(cur)){
      pdebug(__func__, "(%s) failed to load info: %s", cur->name, lm_strerror()); 
      if(!allow_update)
        goto next;
      pdebug(__func__, "(%s) gonna try updating", cur->name); 
    }

    goto success;

  update:
    if(!lm_pool_info_get(cur)){
      pdebug(__func__, "(%s) failed to update info: %s", cur->name, lm_strerror()); 
      goto next;
    }

  success:
    pdebug(__func__, "(%s) loaded pool info", cur->name);
    status = true;

  next:
    if(NULL != callback)
      callback(ctx, cur, status, data);
    status = false;
    cur = cur->next;
  }
}

void lm_ctx_pool_get_list(lm_ctx_t *ctx, bool allow_update, bool force_update, lm_ctx_pool_callback_t callback, void *data){
  lm_pool_t *cur = ctx->pools;
  bool status = false;

  while(NULL != cur){
    if(lm_pool_path_is_empty(cur)){
      pdebug(__func__, "(%s) failed to load list, pool paths are empty", cur->name);
      goto next;
    }

    if(force_update)
      goto update;

    if(!lm_pool_list_load(cur)){
      pdebug(__func__, "(%s) failed to load list: %s", cur->name, lm_strerror()); 
      if(!allow_update)
        goto next;
      pdebug(__func__, "(%s) gonna try updating", cur->name); 
    }
    
    goto success;

  update:
    if(!lm_pool_list_get(cur)){
      pdebug(__func__, "(%s) failed to update list: %s", cur->name, lm_strerror()); 
      goto next;
    }

  success:
    pdebug(__func__, "(%s) loaded pool list", cur->name);
    status = true;

  next:
    if(NULL != callback)
      callback(ctx, cur, status, data);
    status = false;
    cur = cur->next;
  }
}

bool lm_ctx_pool_serve(lm_ctx_t *ctx, char *addr, uint8_t threads) {
  if (NULL == addr || threads < 0) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  struct sockaddr saddr;
  char           *host = NULL;
  lm_mptp_t       packet;
  uint16_t        port;
  int             sock;
  lm_thpool_t     tp;

  lm_thpool_init(&tp, threads);

  if (!parse_host(addr, host, &port))
    return false;

  if (port == 0) {
    lm_error_set(LM_ERR_NoPort);
    return false;
  }

  if ((sock = lm_mptp_server_listen(addr, port)) < 0)
    return false;

  while (lm_mptp_server_recv(sock, &packet, &saddr)) {
    if (!lm_mptp_server_verify(&packet)) {
      pdebug(__func__, "skipping packet, failed to verify: %s", lm_strerror());
      continue;
    }

    char hostname[packet.header.host_size + 1]; // +1 for NULL terminator
    char path[packet.header.data_size + 1], *ppath = path;

    if (!lm_mptp_get_host(&packet, hostname)) {
      pdebug(__func__, "skipping packet, failed to get hostname: %s", lm_strerror());
      continue;
    }

    if (!lm_mptp_get_data(&packet, path)) {
      pdebug(__func__, "skipping packet, failed to get path: %s", lm_strerror());
      continue;
    }

    if(MPTP_FLAGS_CODE(&packet) == MPTP_C2S_PULL && (ppath = dirname(path)) == NULL){
      pdebug(__func__, "skipping packet, failed to get dirname: %s", strerror(errno));
      continue;
    }

    lm_pool_t *pool = lm_ctx_pool_by_url(ctx, hostname, ppath);

    if (NULL == pool) {
      pdebug(__func__, "unknown pool (%s) closing connection", hostname);
      lm_mptp_init(&packet, false, MPTP_S2C_BRUH, true);
      lm_mptp_server_send(sock, &packet, &saddr);
      continue;
    }

    if(lm_pool_path_is_empty(pool)){
      pdebug(__func__, "requested pool (%s) have empty paths, closing connection", pool->name);
      lm_mptp_init(&packet, false, MPTP_S2C_BRUH, true);
      lm_mptp_server_send(sock, &packet, &saddr);
      continue;
    }

    lm_pool_thread_arg_t *arg = malloc(sizeof(lm_pool_thread_arg_t));

    memcpy(&arg->addr, &saddr, sizeof(struct sockaddr));
    lm_mptp_copy(&arg->packet, &packet);
    arg->pool = pool;
    arg->sock = sock;

    pdebug(__func__, "adding new connection to the pool");
    lm_thpool_add(&tp, lm_pool_serve_thread, arg);
  }

  lm_thpool_stop(&tp);
  return true;
}
