#include "../../include/database.h"
#include "../../include/package.h"
#include "../../include/error.h"
#include "../../include/util.h"

#include <sqlite3.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

bool lm_database_add(lm_database_t *db, lm_pkg_t *pkg){
  if(NULL == db || NULL == pkg){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  char depends[lm_package_depend_strlen(pkg)];
  sqlite3_stmt *st = NULL;
  bool ret = false;

  if(sqlite3_prepare(db->sql, queries[QUERY_INSERT_PACKAGE], strlen(queries[QUERY_INSERT_PACKAGE]), &st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for inserting %s: %s", pkg->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(st, 1, pkg->name, strlen(pkg->name), SQLITE_STATIC);
  sqlite3_bind_text(st, 2, pkg->desc, strlen(pkg->desc), SQLITE_STATIC);
  sqlite3_bind_text(st, 3, pkg->version, strlen(pkg->version), SQLITE_STATIC);
  sqlite3_bind_int64(st, 4, pkg->size);

  if(!lm_package_depend_tostr(pkg, depends)){
    pdebug(__func__, "failed to convert depends to string for inserting %s: %s", pkg->name, lm_strerror());
    goto end;
  }
  sqlite3_bind_text(st, 5, depends, strlen(depends), SQLITE_STATIC);

  if(sqlite3_step(st) != SQLITE_DONE){
    pdebug(__func__, "failed to execute insert statement for inserting %s: %s", pkg->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlInsertFail);
    goto end;
  }

  ret = true;
end:
  if(NULL != st)
    sqlite3_finalize(st);
  return ret;
}

bool lm_database_find(lm_database_t *db, lm_pkg_t *pkg, char *name){
  if(NULL == db || NULL == name){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  sqlite3_stmt *st;
  bool ret = false;

  if(sqlite3_prepare(db->sql, queries[QUERY_SELECT_PACKAGE], strlen(queries[QUERY_SELECT_PACKAGE]), &st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for finding %s: %s", name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(st, 1, name, strlen(name), SQLITE_STATIC);

  if(sqlite3_step(st) != SQLITE_ROW){
    pdebug(__func__, "got no rows for %s", name);
    lm_error_set(LM_ERR_DbSqlNotFound);
    goto end;
  }

  if(NULL == pkg){
    ret = true;
    goto end;
  }

  // package pointer should already be free'd
  // we are initing it just in case the caller didn't
  lm_package_init(pkg);

  pkg->name = strdup((char*)sqlite3_column_text(st, 0));
  pkg->desc = strdup((char*)sqlite3_column_text(st, 1));
  pkg->version = strdup((char*)sqlite3_column_text(st, 2));
  pkg->size = sqlite3_column_int64(st, 3);

  char *depends = (char*)sqlite3_column_text(st, 3);
  if(!lm_package_depend_fromstr(pkg, depends)){
    pdebug(__func__, "failed to load depends for finding %s: %s", pkg->name, lm_strerror());
    // error is set by the function
    goto end;
  }
 
  ret = true;
end:
  if(NULL != st)
    sqlite3_finalize(st);
  return ret;
}

bool lm_database_del(lm_database_t *db, lm_pkg_t *pkg){
  if(NULL == db || NULL == pkg){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  sqlite3_stmt *st;
  bool ret = false;

  if(sqlite3_prepare(db->sql, queries[QUERY_DELETE_PACKAGE], strlen(queries[QUERY_DELETE_PACKAGE]), &st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for deleting %s: %s", pkg->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(st, 1, pkg->name, strlen(pkg->name), SQLITE_STATIC);

  if(sqlite3_step(st) != SQLITE_DONE){
    pdebug(__func__, "failed to execute delete statement for deleting %s: %s", pkg->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlInsertFail);
    goto end;
  }

  ret = true;
end:
  if(NULL != st)
    sqlite3_finalize(st);
  return ret;
}
