#include "../../include/package.h"
#include "../../include/error.h"
#include "../../include/pool.h"
#include "../../include/mptp.h"
#include "../../include/util.h"

#include <stdbool.h>
#include <dirent.h>
#include <libgen.h>
#include <string.h>

bool lm_pool_list_load(lm_pool_t *pool){
  if(NULL == pool){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }
  
  lm_pool_list_free(pool);

  if(lm_pool_path_is_empty(pool)){
    lm_error_set(LM_ERR_PoolPathsEmpty);
    return false;
  }

  if(!exists(pool->paths.list_file) || !can_read(pool->paths.list_file)){
    lm_error_set(LM_ERR_PoolListCantRead);
    return false;
  }

  pdebug(__func__, "(%s) extracting pool to %s", pool->name, pool->paths.list_dir);

  if(!mkdir_ifnot(pool->paths.dir)){
    lm_error_set(LM_ERR_PoolBadDir);
    return false;
  }

  if(!mkdir_ifnot(pool->paths.list_dir)){
    lm_error_set(LM_ERR_PoolBadPaths);
    return false;
  }

  size_t ent_len = 0, list_dir_len = strlen(pool->paths.list_dir);
  struct dirent *ent = NULL;
  DIR *dirfd = NULL;
  bool ret = false;

  if(!extract_archive(pool->paths.list_dir, pool->paths.list_file))
    goto end;

  if((dirfd = opendir(pool->paths.list_dir)) == NULL){
    lm_error_set(LM_ERR_PoolListDirFail);
    goto end;
  }
  
  while((ent = readdir(dirfd)) != NULL){
    if(eq(ent->d_name, "..") || eq(ent->d_name, "."))
      continue;

    ent_len = strlen(ent->d_name);
    char datap[ent_len+list_dir_len+10];
    join_multiple(datap, pool->paths.list_dir, ent->d_name, "DATA");
    
    lm_pkg_t *pkg = lm_package_new();
    
    if(!lm_package_data_load(pkg, datap)){
      pdebug(__func__, "(%s) failed to load new package from %s", pool->name, datap);
      lm_package_free(pkg);
      continue;
    }

    if(!lm_pool_package_add(pool, pkg)){
      pdebug(__func__, "(%s) failed to add new package: %s", pool->name, pkg->name);
      lm_package_free(pkg);
      continue;
    }
      
    pdebug(__func__, "(%s) added new package: %s", pool->name, pkg->name);
  }

  ret = true;

end:
  if(NULL != dirfd)
    closedir(dirfd);
  return ret;
}

bool lm_pool_list_get(lm_pool_t *pool) {
  if(NULL == pool){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(!pool->available){
    lm_error_set(LM_ERR_PoolNotAvailable);
    return false;
  }

  if(lm_pool_path_is_empty(pool)){
    lm_error_set(LM_ERR_PoolPathsEmpty);
    return false;
  }

  if(!mkdir_ifnot(pool->paths.dir)){
    lm_error_set(LM_ERR_PoolBadDir);
    return false;
  }

  if(!mkdir_ifnot(pool->paths.list_dir)){
    lm_error_set(LM_ERR_PoolBadPaths);
    return false;
  }

  if(NULL == pool->url.path || NULL == pool->url.host){
    lm_error_set(LM_ERR_PoolUrlEmpty);
    return false;
  }

  int sock = -1;
  lm_mptp_t packet;
  bool ret = false;

  if((sock = lm_mptp_client_connect(pool->url.host, pool->url.port)) < 0)
    return false;

  lm_mptp_init(&packet, true, MPTP_C2S_LIST, true);
  lm_mptp_set_host(&packet, pool->url.host);
  lm_mptp_set_data(&packet, pool->url.path, strlen(pool->url.path));

  if(!lm_mptp_client_send(sock, &packet))
    goto end;
 
  if(!lm_mptp_recvfile(sock, pool->paths.list_file))
    goto end;

  ret = true;
end:
  lm_mptp_close(sock);
  if(ret)
    ret = lm_pool_list_load(pool);
  return ret;
}

void lm_pool_list_free(lm_pool_t *pool){
  if(NULL == pool)
    return;

  lm_pkg_t *cur = pool->pkg, *prev = NULL;
  while(NULL != cur){
    prev = cur;
    cur = cur->next;
    lm_package_free(prev);
  }
}
