#include "../../include/error.h"
#include "../../include/util.h"
#include "../../include/pool.h"

#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>

void lm_pool_serve(lm_pool_t *pool, lm_mptp_t *packet, int sock, struct sockaddr *addr) {
  char ipaddr[INET6_ADDRSTRLEN];
  bzero(ipaddr, sizeof(ipaddr));
  sockaddr_to_str(addr, ipaddr);

  switch (MPTP_FLAGS_CODE(packet)) {

  // response PING with PONG
  case MPTP_C2S_PING:
    pdebug(__func__, "(%s) PING from %s returning PONG", pool->name, ipaddr);
    lm_mptp_init(packet, false, MPTP_S2C_PONG, true);
    lm_mptp_server_send(sock, packet, addr);
    break;

  // when INFO file is requested, send the file
  case MPTP_C2S_INFO:
    pdebug(__func__, "(%s) INFO from %s attempting to send info", pool->name, ipaddr);
    lm_mptp_sendfile(sock, addr, pool->paths.info_file);
    break;

  // when LIST file is requested, send the file
  case MPTP_C2S_LIST:
    pdebug(__func__, "(%s) LIST from %s attempting to send list", pool->name, ipaddr);
    lm_mptp_sendfile(sock, addr, pool->paths.list_file);
    break;
 
  // when the request code is PULL, send the requested package archive and
  // requested package signature
  case MPTP_C2S_PULL:
    // PULL request should contain package path,
    // if path (stored in the data field) is empty it's an invalid request
    if(packet->header.data_size <= 0){
      lm_mptp_init(packet, false, MPTP_S2C_WHAT, true);
      lm_mptp_server_send(sock, packet, addr);
      break;
    }
    
    pdebug(__func__, "(%s) PULL from %s attempting to send package archive and signature", pool->name, ipaddr);
    char path[packet->header.data_size + 1], *package = path;

    if(!lm_mptp_get_data(packet, path)){
      // we should never be able to get here, if we do theres definetly a bug
      pdebug(__func__, "(%s) skipping PULL from %s, failed to get path: %s", pool->name, ipaddr, lm_strerror());
      break;
    }

    // if we can't get the package name, then theres something wrong with the request
    if((package = basename(path)) == NULL){
      lm_mptp_init(packet, false, MPTP_S2C_WHAT, true);
      lm_mptp_server_send(sock, packet, addr);
      break;
    }

    size_t package_size = strlen(package);
    char name[package_size+1], version[package_size+1]; 
    lm_pkg_t *pkg = NULL;

    // if we can't parse the package name, request is invalid
    if(!package_parse(package, name, version)){
      lm_mptp_init(packet, false, MPTP_S2C_WHAT, true);
      lm_mptp_server_send(sock, packet, addr);
      break;
    }

    // if the package is not found in the pool, tell the client
    if((pkg = lm_pool_package_find(pool, name, version)) == NULL){
      lm_mptp_init(packet, false, MPTP_S2C_BRUH, true);
      lm_mptp_server_send(sock, packet, addr);
      break;
    }

    // send package archive and the signature file
    lm_mptp_sendfile(sock, addr, pkg->paths.archive);
    lm_mptp_sendfile(sock, addr, pkg->paths.signature);
    break;
  }
}

void lm_pool_serve_thread(void *_arg) {
  lm_pool_thread_arg_t *arg = _arg;
  lm_pool_serve(arg->pool, &arg->packet, arg->sock, &arg->addr);
  free(arg);
}
